@@ -38,8 +38,7 @@
 #include "version.h"
 #include "header.h"
 
-extern int verbose;
-extern int quiet;
+extern int analyze_cache(cdrom_drive *d, FILE *progress, FILE *log, int speed);
 
 static long parse_offset(cdrom_drive *d, char *offset, int begin){
   long track=-1;
@@ -72,9 +71,7 @@ static long parse_offset(cdrom_drive *d, char *offset, int begin){
       offset[chars]='\0';
       track=atoi(offset);
       if(track<0 || track>d->tracks){ /*take track 0 as pre-gap of 1st track*/
-	char buffer[256];
-	sprintf(buffer,"Track #%ld does not exist.",track);
-	report(buffer);
+	report("Track #%ld does not exist.",track);
 	exit(1);
       }
     }
@@ -174,31 +171,24 @@ static void display_toc(cdrom_drive *d){
 	 "===========================================================");
   
   for(i=1;i<=d->tracks;i++)
-    if(cdda_track_audiop(d,i)){
-      char buffer[256];
+    if(cdda_track_audiop(d,i)>0){
 
       long sec=cdda_track_firstsector(d,i);
       long off=cdda_track_lastsector(d,i)-sec+1;
       
-      sprintf(buffer,
-	      "%3d.  %7ld [%02d:%02d.%02d]  %7ld [%02d:%02d.%02d]  %s %s %s",
-	      i,
-	      off,(int)(off/(60*75)),(int)((off/75)%60),(int)(off%75),
-	      sec,(int)(sec/(60*75)),(int)((sec/75)%60),(int)(sec%75),
-	      cdda_track_copyp(d,i)?"  OK":"  no",
-	      cdda_track_preemp(d,i)?" yes":"  no",
-	      cdda_track_channels(d,i)==2?" 2":" 4");
-      report(buffer);
+      report("%3d.  %7ld [%02d:%02d.%02d]  %7ld [%02d:%02d.%02d]  %s %s %s",
+	     i,
+	     off,(int)(off/(60*75)),(int)((off/75)%60),(int)(off%75),
+	     sec,(int)(sec/(60*75)),(int)((sec/75)%60),(int)(sec%75),
+	     cdda_track_copyp(d,i)?"  OK":"  no",
+	     cdda_track_preemp(d,i)?" yes":"  no",
+	     cdda_track_channels(d,i)==2?" 2":" 4");
       audiolen+=off;
     }
-  {
-    char buffer[256];
-    sprintf(buffer, "TOTAL %7ld [%02d:%02d.%02d]    (audio only)",
-	    audiolen,(int)(audiolen/(60*75)),(int)((audiolen/75)%60),
-	    (int)(audiolen%75));
-      report(buffer);
-  }
-  report("");
+  report("TOTAL %7ld [%02d:%02d.%02d]    (audio only)",
+	 audiolen,(int)(audiolen/(60*75)),(int)((audiolen/75)%60),
+	 (int)(audiolen%75));
+  report(" ");
 }
 
 static void usage(FILE *f){
@@ -214,11 +204,20 @@ VERSION"\n"
 "  cdparanoia [options] <span> [outfile]\n\n"
 
 "OPTIONS:\n"
+"  -A --analyze-drive              : run and log a complete analysis of drive\n"
+"                                    caching, timing and reading behavior;\n"
+"                                    verifies that cdparanoia is correctly\n"
+"                                    modelling a sprcific drive's cache and\n"
+"                                    read behavior. Implies -vQL\n\n"
 "  -v --verbose                    : extra verbose operation\n"
 "  -q --quiet                      : quiet operation\n"
 "  -e --stderr-progress            : force output of progress information to\n"
 "                                    stderr (for wrapper scripts)\n"
-"  -l --log-summary         <file> : save result summary to file\n"
+"  -l --log-summary [<file>]       : save result summary to file, default\n"
+"                                    filename cdparanoia.log\n"
+"  -L --log-debug   [<file>]       : save detailed device autosense and\n"
+"                                    debugging output to file, default\n"
+"                                    filename cdparanoia.log\n"
 "  -V --version                    : print version info and quit\n"
 "  -Q --query                      : autosense drive, query disc and quit\n"
 "  -B --batch                      : 'batch' mode (saves each track to a\n"
@@ -242,8 +241,16 @@ VERSION"\n"
 "                                    verification to n sectors\n"
 "  -d --force-cdrom-device   <dev> : use specified device; disallow \n"
 "                                    autosense\n"
-"  -g --force-generic-device <dev> : use specified generic scsi device\n"
-"  -S --force-read-speed <n>       : read from device at specified speed\n"
+"  -k --force-cooked-device  <dev> : use specified cdrom device and force\n"
+"                                    use of the old 'cooked ioctl' kernel\n"
+"                                    interface. -k cannot be used with -d\n"
+"                                    or -g.\n"
+"  -g --force-generic-device <dev> : use specified generic scsi device and\n"
+"                                    force use of the old SG kernel\n"
+"                                    interface. -g cannot be used with -k.\n"
+"  -S --force-read-speed <n>       : read from device at specified speed; by\n"
+"                                    default, cdparanoia sets drive to full\n"
+"                                    speed.\n"
 "  -t --toc-offset <n>             : Add <n> sectors to the values reported\n"
 "                                    when addressing tracks. May be negative\n"
 "  -T --toc-bias                   : Assume that the beginning offset of \n"
@@ -332,7 +339,7 @@ long callbegin;
 long callend;
 long callscript=0;
 
-static char *callback_strings[15]={"wrote",
+static char *callback_strings[16]={"wrote",
                                    "finished",
 				   "read",
 				   "verify",
@@ -346,7 +353,8 @@ static char *callback_strings[15]={"wrote",
 				   "overlap",
 				   "dropped",
 				   "duped",
-				   "transport error"};
+				   "transport error",
+                                   "cache error"};
 
 static int skipped_flag=0;
 static int abort_on_skip=0;
@@ -374,12 +382,27 @@ static void callback(long inpos, int function){
   static int slevel=0;
   static int slast=0;
   static int stimeout=0;
+  static int cacheerr=0;
   char *smilie="= :-)";
   
   if(callscript)
     fprintf(stderr,"##: %d [%s] @ %ld\n",
 	    function,(function>=-2&&function<=13?callback_strings[function+2]:
 		      ""),inpos);
+  else{
+    if(function==PARANOIA_CB_CACHEERR){
+      if(!cacheerr){
+	fprintf(stderr,
+		"\rWARNING: The CDROM drive appears to be seeking impossibly quickly.\n"
+		"This could be due to timer bugs, a drive that really is improbably fast,\n"
+		"or, most likely, a bug in cdparanoia's cache modelling.\n\n"
+		"Please consider using the -A option to perform an analysis run, then mail\n"
+		"the cdparanoia.log file produced by the analysis to paranoia-dev@xiph.org\n"
+		"to assist developers in correcting the problem.\n\n");
+      }
+      cacheerr++;
+    }
+  }
 
   if(!quiet){
     long test;
@@ -444,12 +467,17 @@ static void callback(long inpos, int function){
 	    break;
 	  case PARANOIA_CB_READERR:
 	    slevel=6;
-	    if(dispcache[position]!='V')
+	    if(dispcache[position]!='V' && dispcache[position]!='C')
 	      dispcache[position]='e';
 	    break;
+	  case PARANOIA_CB_CACHEERR:
+	    slevel=8;
+	    dispcache[position]='C';
+	    break;
 	  case PARANOIA_CB_SKIP:
 	    slevel=8;
-	    dispcache[position]='V';
+	    if(dispcache[position]!='C')
+	      dispcache[position]='V';
 	    break;
 	  case PARANOIA_CB_OVERLAP:
 	    overlap=osector;
@@ -576,7 +604,7 @@ static void callback(long inpos, int function){
     memset(dispcache,' ',graph);
 }
 
-const char *optstring = "escCn:o:O:d:g:S:prRwafvqVQhZz::YXWBi:Tt:l:";
+const char *optstring = "escCn:o:O:d:g:k:S:prRwafvqVQhZz::YXWBi:Tt:l::L::A";
 
 struct option options [] = {
 	{"stderr-progress",no_argument,NULL,'e'},
@@ -586,6 +614,7 @@ struct option options [] = {
 	{"force-default-sectors",required_argument,NULL,'n'},
 	{"force-search-overlap",required_argument,NULL,'o'},
 	{"force-cdrom-device",required_argument,NULL,'d'},
+	{"force-cooked-device",required_argument,NULL,'k'},
 	{"force-generic-device",required_argument,NULL,'g'},
 	{"force-read-speed",required_argument,NULL,'S'},
 	{"sample-offset",required_argument,NULL,'O'},
@@ -603,13 +632,15 @@ struct option options [] = {
 	{"version",no_argument,NULL,'V'},
 	{"query",no_argument,NULL,'Q'},
 	{"help",no_argument,NULL,'h'},
+	{"analyze-drive",no_argument,NULL,'A'},
 	{"disable-paranoia",no_argument,NULL,'Z'},
 	{"disable-extra-paranoia",no_argument,NULL,'Y'},
 	{"abort-on-skip",no_argument,NULL,'X'},
 	{"disable-fragmentation",no_argument,NULL,'F'},
 	{"output-info",required_argument,NULL,'i'},
 	{"never-skip",optional_argument,NULL,'z'},
-	{"log-summary",required_argument,NULL,'l'},
+	{"log-summary",optional_argument,NULL,'l'},
+	{"log-debug",optional_argument,NULL,'L'},
 
 	{NULL,0,NULL,0}
 };
@@ -646,13 +677,20 @@ int main(int argc,char *argv[]){
   int force_cdrom_overlap=-1;
   char *force_cdrom_device=NULL;
   char *force_generic_device=NULL;
-  int force_cdrom_speed=-1;
+  char *force_cooked_device=NULL;
+  int force_cdrom_speed=0;
   int max_retries=20;
   char *span=NULL;
   int output_type=1; /* 0=raw, 1=wav, 2=aifc */
   int output_endian=0; /* -1=host, 0=little, 1=big */
   int query_only=0;
   int batch=0,i;
+  int run_cache_test=0;
+
+  char *logfile_name=NULL;
+  char *reportfile_name=NULL;
+  int logfile_open=0;
+  int reportfile_open=0;
 
   /* full paranoia, but allow skipping */
   int paranoia_mode=PARANOIA_MODE_FULL^PARANOIA_MODE_NEVERSKIP; 
@@ -687,9 +725,22 @@ int main(int argc,char *argv[]){
       force_cdrom_device=copystring(optarg);
       break;
     case 'g':
+      if(force_cooked_device){
+	report("-g option incompatable with -k\n");
+	exit(1);
+      }
+      force_cooked_device=NULL;
       if(force_generic_device)free(force_generic_device);
       force_generic_device=copystring(optarg);
       break;
+    case 'k':
+      if(force_generic_device || force_cdrom_device){
+	report("-k option incompatable with -d and -g\n");
+	exit(1);
+      }
+      if(force_cooked_device)free(force_cooked_device);
+      force_cooked_device=copystring(optarg);
+      break;
     case 'S':
       force_cdrom_speed=atoi(optarg);
       break;
@@ -730,7 +781,7 @@ int main(int argc,char *argv[]){
       break;
     case 'e':
       callscript=1;
-      fprintf(stderr,"Sending all callcaks to stderr for wrapper script\n");
+      fprintf(stderr,"Sending all callbacks to stderr for wrapper script\n");
       break;
     case 'V':
       fprintf(stderr,VERSION);
@@ -746,6 +797,12 @@ int main(int argc,char *argv[]){
     case 'Z':
       paranoia_mode=PARANOIA_MODE_DISABLE; 
       break;
+    case 'A':
+      run_cache_test=1;
+      query_only=1;
+      reportfile_open=1;
+      verbose=CDDA_MESSAGE_PRINTIT;
+      break;
     case 'z':
       if (optarg) {
         max_retries = atoi (optarg);
@@ -780,18 +837,18 @@ int main(int argc,char *argv[]){
       toc_offset=atoi(optarg);
       break;
     case 'l':
-      if(logfile && logfile != stdout)fclose(logfile);
-      if(!strcmp(optarg,"-"))
-	logfile=stdout;
-      else{
-	logfile=fopen(optarg,"w");
-	if(logfile==NULL){
-	  report3("Cannot open log summary file %s: %s",(char*)optarg,
-		  strerror(errno));
-	  exit(1);
-	}
-      }
-
+      if(logfile_name)free(logfile_name);
+      logfile_name=NULL;
+      if(optarg)
+	logfile_name=strdup(optarg);
+      logfile_open=1;
+      break;
+    case 'L':
+      if(reportfile_name)free(reportfile_name);
+      reportfile_name=NULL;
+      if(optarg)
+	reportfile_name=strdup(optarg);
+      reportfile_open=1;
       break;
     case 'O':
       sample_offset=atoi(optarg);
@@ -802,18 +859,67 @@ int main(int argc,char *argv[]){
     }
   }
 
+  if(logfile_open){
+    if(logfile_name==NULL)
+      logfile_name=strdup("cdparanoia.log");
+    if(!strcmp(logfile_name,"-")){
+      logfile=stdout;
+      logfile_open=0;
+    }else{
+      logfile=fopen(logfile_name,"w");
+      if(logfile==NULL){
+	report("Cannot open log summary file %s: %s",logfile_name,
+	       strerror(errno));
+	exit(1);
+      }
+    }
+  }
+  if(reportfile_open){
+    if(reportfile_name==NULL)
+      reportfile_name=strdup("cdparanoia.log");
+    if(!strcmp(reportfile_name,"-")){
+      reportfile=stdout;
+      reportfile_open=0;
+    }else{
+      if(logfile_name && !strcmp(reportfile_name,logfile_name)){
+	reportfile=logfile;
+	reportfile_open=0;
+      }else{
+	reportfile=fopen(reportfile_name,"w");
+	if(reportfile==NULL){
+	  report("Cannot open debug log file %s: %s",reportfile_name,
+		 strerror(errno));
+	  exit(1);
+	}
+      }
+    }
+  }
+    
   if(logfile){
     /* log command line and version */
     int i;
     for (i = 0; i < argc; i++) 
       fprintf(logfile,"%s ",argv[i]);
     fprintf(logfile,"\n");
-    
-    fprintf(logfile,VERSION);
-    fprintf(logfile,"\n");
+
+    if(reportfile!=logfile){
+      fprintf(logfile,VERSION);
+      fprintf(logfile,"\n");
+      fprintf(logfile,"Using cdda library version: %s\n",cdda_version());
+      fprintf(logfile,"Using paranoia library version: %s\n",paranoia_version());
+    }
     fflush(logfile);
   }
 
+  if(reportfile && reportfile!=logfile){
+    /* log command line */
+    int i;
+    for (i = 0; i < argc; i++) 
+      fprintf(reportfile,"%s ",argv[i]);
+    fprintf(reportfile,"\n");
+    fflush(reportfile);
+  }
+
   if(optind>=argc && !query_only){
     if(batch)
       span=NULL;
@@ -826,10 +932,16 @@ int main(int argc,char *argv[]){
     span=copystring(argv[optind]);
 
   report(VERSION);
+  if(verbose){
+    report("Using cdda library version: %s",cdda_version());
+    report("Using paranoia library version: %s",paranoia_version());
+  }
 
   /* Query the cdrom/disc; we may need to override some settings */
 
-  if(force_generic_device)
+  if(force_cooked_device){
+    d=cdda_identify_cooked(force_cooked_device,verbose,NULL);
+  }else if(force_generic_device)
     d=cdda_identify_scsi(force_generic_device,force_cdrom_device,verbose,NULL);
   else
     if(force_cdrom_device)
@@ -856,7 +968,7 @@ int main(int argc,char *argv[]){
 	    report("\n");
 	    exit(1);
 	  }else
-	    report("");
+	    report(" ");
 	}
       }
 
@@ -890,14 +1002,10 @@ int main(int argc,char *argv[]){
       d=NULL;
       exit(1);
     }
-    {
-      char buffer[256];
-      sprintf(buffer,"Forcing default to read %d sectors; "
-	      "ignoring preset and autosense",force_cdrom_sectors);
-      report(buffer);
-      d->nsectors=force_cdrom_sectors;
-      d->bigbuff=force_cdrom_sectors*CD_FRAMESIZE_RAW;
-    }
+    report("Forcing default to read %d sectors; "
+	   "ignoring preset and autosense",force_cdrom_sectors);
+    d->nsectors=force_cdrom_sectors;
+    d->bigbuff=force_cdrom_sectors*CD_FRAMESIZE_RAW;
   }
   if(force_cdrom_overlap!=-1){
     if(force_cdrom_overlap<0 || force_cdrom_overlap>75){
@@ -906,12 +1014,8 @@ int main(int argc,char *argv[]){
       d=NULL;
       exit(1);
     }
-    {
-      char buffer[256];
-      sprintf(buffer,"Forcing search overlap to %d sectors; "
-	      "ignoring autosense",force_cdrom_overlap);
-      report(buffer);
-    }
+    report("Forcing search overlap to %d sectors; "
+	   "ignoring autosense",force_cdrom_overlap);
   }
 
   switch(cdda_open(d)){
@@ -928,6 +1032,44 @@ int main(int argc,char *argv[]){
     exit(1);
   }
 
+  if(force_cdrom_speed==0)force_cdrom_speed=-1;
+  if(force_cdrom_speed!=-1){
+    report("\nAttempting to set speed to %dx... ",force_cdrom_speed);
+  }else{
+    if(verbose)
+      report("\nAttempting to set cdrom to full speed... ");
+  }
+
+  if(cdda_speed_set(d,force_cdrom_speed)){
+    if(verbose || force_cdrom_speed!=-1)
+      report("\tCDROM speed set FAILED. Continuing anyway...");
+  }else{
+    if(verbose)
+      report("\tdrive returned OK.");
+  }
+  
+  if(run_cache_test){
+    int warn=analyze_cache(d, stderr, reportfile, force_cdrom_speed);
+    
+    if(warn==0){
+      reportC("\nDrive tests OK with Paranoia.\n\n");
+      return 0;
+    }
+
+    if(warn==1)
+      reportC("\nWARNING! PARANOIA MAY NOT BE TRUSTWORTHY WITH THIS DRIVE!\n"
+	      "\nThe Paranoia library may not model this CDROM drive's cache"
+	      "\ncorrectly according to this analysis run. Analysis is not"
+	      "\nalways accurate (it can be fooled by machine load or random"
+	      "\nkernel latencies), but if a failed result happens more often"
+	      "\nthan one time in twenty on an unloaded machine, please mail"
+	      "\nthe %s file produced by this failed analysis to"
+	      "\nparanoia-dev@xiph.org to assist developers in extending"
+	      "\nParanoia to handle this CDROM properly.\n\n",reportfile_name);
+    return 1;
+  }
+
+
   /* Dump the TOC */
   if(query_only || verbose)display_toc(d);
   if(query_only)exit(0);
@@ -952,10 +1094,6 @@ int main(int argc,char *argv[]){
     d->disc_toc[i].dwStartSector+=toc_offset;
 
 
-  if(force_cdrom_speed!=-1){
-    cdda_speed_set(d,force_cdrom_speed);
-  }
-
   if(d->nsectors==1){
     report("WARNING: The autosensed/selected sectors per read value is\n"
 	   "         one sector, making it very unlikely Paranoia can \n"
@@ -1018,7 +1156,6 @@ int main(int argc,char *argv[]){
     }
 
     {
-      char buffer[250];
       int track1=cdda_sector_gettrack(d,first_sector);
       int track2=cdda_sector_gettrack(d,last_sector);
       long off1=first_sector-cdda_track_firstsector(d,track1);
@@ -1031,12 +1168,11 @@ int main(int argc,char *argv[]){
 	  exit(1);
 	}
 
-      sprintf(buffer,"Ripping from sector %7ld (track %2d [%d:%02d.%02d])\n"
-	      "\t  to sector %7ld (track %2d [%d:%02d.%02d])\n",first_sector,
-	      track1,(int)(off1/(60*75)),(int)((off1/75)%60),(int)(off1%75),
-	      last_sector,
-	      track2,(int)(off2/(60*75)),(int)((off2/75)%60),(int)(off2%75));
-      report(buffer);
+      report("Ripping from sector %7ld (track %2d [%d:%02d.%02d])\n"
+	     "\t  to sector %7ld (track %2d [%d:%02d.%02d])\n",first_sector,
+	     track1,(int)(off1/(60*75)),(int)((off1/75)%60),(int)(off1%75),
+	     last_sector,
+	     track2,(int)(off2/(60*75)),(int)((off2/75)%60),(int)(off2%75));
       
     }
 
@@ -1135,13 +1271,13 @@ int main(int argc,char *argv[]){
 	    
 	    out=open(outfile_name,O_RDWR|O_CREAT|O_TRUNC,0666);
 	    if(out==-1){
-	      report3("Cannot open specified output file %s: %s",outfile_name,
+	      report("Cannot open specified output file %s: %s",outfile_name,
 		      strerror(errno));
 	      cdda_close(d);
 	      d=NULL;
 	      exit(1);
 	    }
-	    report2("outputting to %s\n",outfile_name);
+	    report("outputting to %s\n",outfile_name);
 	    if(logfile){
 	      fprintf(logfile,"outputting to %s\n",outfile_name);
 	      fflush(logfile);
@@ -1171,13 +1307,13 @@ int main(int argc,char *argv[]){
 	  
 	  out=open(outfile_name,O_RDWR|O_CREAT|O_TRUNC,0666);
 	  if(out==-1){
-	    report3("Cannot open default output file %s: %s",outfile_name,
+	    report("Cannot open default output file %s: %s",outfile_name,
 		    strerror(errno));
 	    cdda_close(d);
 	    d=NULL;
 	    exit(1);
 	  }
-	  report2("outputting to %s\n",outfile_name);
+	  report("outputting to %s\n",outfile_name);
 	  if(logfile){
 	    fprintf(logfile,"outputting to %s\n",outfile_name);
 	    fflush(logfile);
@@ -1206,7 +1342,7 @@ int main(int argc,char *argv[]){
 	  if(buffering_write(out,
 			     ((char *)offset_buffer)+offset_buffer_used,
 			     CD_FRAMESIZE_RAW-offset_buffer_used)){
-	    report2("Error writing output: %s",strerror(errno));
+	    report("Error writing output: %s",strerror(errno));
 	    exit(1);
 	  }
 	}
@@ -1226,6 +1362,10 @@ int main(int argc,char *argv[]){
 	  if(err)free(err);
 	  if(mes)free(mes);
 	  if(readbuf==NULL){
+	    if(errno==EBADF || errno==ENOMEDIUM){
+	      report("\nparanoia_read: CDROM drive unavailable, bailing.\n");
+	      exit(1);
+	    }
 	    skipped_flag=1;
 	    report("\nparanoia_read: Unrecoverable error, bailing.\n");
 	    break;
@@ -1247,7 +1387,7 @@ int main(int argc,char *argv[]){
 
 	  if(buffering_write(out,((char *)readbuf)+offset_skip,
 			     CD_FRAMESIZE_RAW-offset_skip)){
-	    report2("Error writing output: %s",strerror(errno));
+	    report("Error writing output: %s",strerror(errno));
 	    exit(1);
 	  }
 	  offset_skip=0;
@@ -1292,7 +1432,7 @@ int main(int argc,char *argv[]){
 
 	    if(buffering_write(out,(char *)offset_buffer,
 			       offset_buffer_used)){
-	      report2("Error writing output: %s",strerror(errno));
+	      report("Error writing output: %s",strerror(errno));
 	      exit(1);
 	    }
 	  }
@@ -1301,7 +1441,7 @@ int main(int argc,char *argv[]){
 	buffering_close(out);
 	if(skipped_flag){
 	  /* remove the file */
-	  report2("\nRemoving aborted file: %s",outfile_name);
+	  report("\nRemoving aborted file: %s",outfile_name);
 	  unlink(outfile_name);
 	  /* make the cursor correct if we have another track */
 	  if(batch_track!=-1){
@@ -1324,7 +1464,9 @@ int main(int argc,char *argv[]){
   
   cdda_close(d);
   d=NULL;
-  if(logfile && logfile != stdout)
+  if(logfile_open)
     fclose(logfile);
+  if(reportfile_open)
+    fclose(reportfile);
   return 0;
 }
